import os
from time import sleep
from datetime import datetime
from . import utils
from .models import WeirdGitterResult
import tomllib


class WeirdGitter:
    """
    Class for git operations
    Checks if repo exists, else clones it. Switches to branch, then with given timeout check if branch is up to date
    If pull - run wg_test.sh, if ["TEST DONE OK"] in res - run wg_build.sh, then wg_deploy.sh
    
    :attribute repo_path: path to repo (it is better to be absolute)
    :attribute branch: name of branch
    :attribute repo_url: url of repo [Optional]
    :attribute timeout: timeout in seconds [Default=300]
    :attribute callback: callback function to call after run
        Callback function should take one arg - wg_result. See WeirdGitterResult for info
    :attribute run_on_start: bool whether to run on start or on pull only
    :attribute debug: bool whether to callback on startup and sys events
    :attribute debug_empty: bool whether to callback on "empty" runs
    :attribute: wgtoml: path to wgtoml
    """
    def __init__(
        self,
        repo_path: str,
        branch: str = 'master',
        name: str = 'WeirdGitter',
        repo_url: str = None,
        timeout: int = 300,
        callback: callable = None,
        run_on_start: bool = True,
        debug=False,
        debug_empty=False,
        wgtoml='weirdgitter.toml'
    ):
        self.repo_path = repo_path
        self.branch = branch
        self.repo_url = repo_url
        self.name = name
        self.timeout = timeout
        if callable(callback):
            self.callback = callback
        else:
            self.callback = WeirdGitterResult.print
        self.run_on_start = run_on_start
        self.debug = debug
        self.debug_empty = debug_empty
        
        self.config = tomllib.loads(open(os.path.join(self.repo_path,wgtoml), 'r').read())
        
        self.__init_repo__()
        
    def __init_repo__(self):
        # Check for path or create
        if not os.path.exists(self.repo_path):
            os.makedirs(self.repo_path)
            
        # Check if repo exists else clone
        status = utils.exec_cmd('git status', cwd=self.repo_path)
        if status['code'] != 0:
            self.callback(self.__get_result__(status='Failed to init repo', error=status['stderr'], msg=status['stdout'] + "\nCloning..."))
            
            #Cloning
            if not self.repo_url:
                raise ValueError('No repo url provided and no git repo in repo path')
            stat = utils.exec_cmd(f'git clone {self.repo_url} .', cwd=self.repo_path)
            if stat['code'] != 0:
                self.callback(self.__get_result__(status='Failed to clone repo', error=stat['stderr'], msg=stat['stdout']))
                raise SystemExit('Failure at clone')
            raise SystemExit('Failure at init_repo')
        
        # Switch to branch
        stat = utils.exec_cmd(f'git checkout -f {self.branch}', cwd=self.repo_path)
        if stat['code'] != 0:
            self.callback(self.__get_result__(status='Failed to switch to branch', error=stat['stderr'], msg=stat['stdout']))
            raise SystemExit('Failure at switch')
        
        if self.debug:
            self.callback(self.__get_result__(status='Repo initialized'))
        
    def __get_result__(
        self,
        status=None,
        error=None,
        test_result=None,
        build_result=None,
        deploy_result=None,
        **kwargs
    ):
        kwargs['result_status'] = status
        kwargs['error'] = error
        kwargs['test_result'] = test_result
        kwargs['build_result'] = build_result
        kwargs['deploy_result'] = deploy_result
        
        updates = [k for k in kwargs.keys() if kwargs[k]]
        
        kwargs['repo_path'] = self.repo_path
        kwargs['name'] = self.name
        kwargs['branch'] = self.branch
        kwargs['repo_url'] = self.repo_url
        kwargs['dt'] = datetime.now().strftime('%c')
        kwargs['updates'] = updates
        
        return WeirdGitterResult(**kwargs)
    
    def run(self):
        """
        Run scripts in root folder
        """
        win = os.name == 'nt'
        if not win and os.name != 'posix':
            raise NotImplementedError(f'OS {os.name} not supported')
        
        stages = self.config['stages']
        
        # Check existence
        res = os.listdir(self.repo_path)
        for i in stages:
            if i not in res:
                self.callback(self.__get_result__(error=f'No {i} in repo', status='ERROR'))
                raise SystemExit('Failure at check existence')
                
        if len(stages) == 0:
            self.callback(self.__get_result__(error='No files to run', status='ERROR'))
        
        logs = {}
        prefix = './' if not win else ''
        for i in stages:
            res = utils.exec_cmd(f'{prefix}{i}', cwd=self.repo_path)
            if res['code'] != 0:
                self.callback(self.__get_result__(error=res['stderr'], status='ERROR'))
                break
            else:
                logs[i] = res['stdout']
        if len(logs) > 0:
            self.callback(self.__get_result__(**logs, status='SUCCESS'))
        else:
            self.callback(self.__get_result__(status='NO DATA', msg='Run done but no data was output by scripts'))
    
    def check_updates(self):
        """
        Git pull, then run scripts if any changes were pulled
        """
        stat = utils.exec_cmd(f'git pull -f', cwd=self.repo_path)
        if stat['code'] != 0:
            self.callback(self.__get_result__(status='Failed to pull repo', error=stat['stderr'], msg=stat['stdout']))
            raise SystemExit('Failure at check_updates')
        if 'Already up to date' in stat['stdout']:
            if self.debug_empty:
                self.callback(self.__get_result__(msg='No updates'))
            else:
                pass
        else:
            print(stat['stdout'])
            self.run()
    
    def run_loop(
        self,
        ignore_errors=False
        ):
        """
        While True loop! Loop for infinite git pull and run scripts by self.timeout
        
        :param ignore_errors" bool
        """
        if self.run_on_start:
            self.run()
        
        while True:
            try:
                self.check_updates()
                sleep(self.timeout)
            except Exception as e:
                if not ignore_errors:
                    raise e
                self.callback(self.__get_result__(error=str(e), status='ERROR'))
            


import chardet
import subprocess


def decode_auto(raw: bytes) -> str:
    """
    :param raw: bytes
    :return: str
    
    Automatically detects encoding and returns decoded string
    """
    detector = chardet.universaldetector.UniversalDetector()
    detector.feed(raw)
    detector.close()
    return raw.decode(detector.result['encoding'] or 'utf-8')


def exec_cmd(cmd: str, cwd: str = None) -> dict:
    """
    :param cmd: str Command to execute
    :param cwd: str Directory to run in
    :return: dict {'code': int, 'stdout': str, 'stderr': str}
    
    Run command and return outputs and code
    """
    p = subprocess.run(cmd, shell=True, capture_output=True, cwd=cwd)
    return {
        'stdout': decode_auto(p.stdout),
        'stderr': decode_auto(p.stderr),
        'code': p.returncode
    }
    
    




class WeirdGitterResult:
    """
    Result of a WeirdGitter run
    Attributes:
        repo_path (str): Path to the repo
        name (str): Name of the WeirdGitter instance
        branch (str): Branch of the repo
        repo_url (str): URL of the repo
        result_status (str): Status of the result (None if no updates, Failed, OK, generic)
        error (str): Error message if result_status is Failed
        test_result (str): Result of running wg_test.sh
        build_result (str): Result of running wg_build.sh
        deploy_result (str): Result of running wg_deploy.sh
        updates (list): List of updates made to the repo
    """
    def __init__(self, **kwargs):
        self.__dict__ = {
            'repo_path': None,
            'name': None,
            'branch': None,
            'repo_url': None,
            'result_status': None, # Failed, OK, or generic, None if no updates
            'error': None, # Error message (str) or None
            'test_result': None,
            'build_result': None,
            'deploy_result': None, # All of the results are text, containing full stdout/stderr of running stages
            'updates': []
        }
        self.__dict__.update(kwargs)
        self.result_status = kwargs.get('result_status', None)
        
    @staticmethod
    def print(wgr: 'WeirdGitterResult'):
        if wgr.result_status is None:
            print(f" {wgr.dt} - {wgr.name} | Nothing changed ".center(100, '-'))
            return
        print(f"{wgr.name}".center(100, '-'))
        for k in wgr.updates:
            print(f"Update in `{k}`:\n\t{wgr.__dict__[k]}")
        print('='*100)